#ifndef KBF_TASK_H
#define KBF_TASK_H

#include <string>
#include <vector>
#include <memory>
#include <utility>

#include <freertos/FreeRTOS.h>
#include <freertos/task.h>

#include "kbf/exception.h"

/**
 * @brief Task handling functions.
 */
namespace kbf::task {
    /** @brief Tag used for logging. */
    static constexpr const char *const TAG = "kbf::task";

    class Task;

    template<class T>
    std::shared_ptr<Task>
    start(const std::string &name, void *arg, uint32_t stackSize = 2048, uint32_t priority = 5);

    /**
     * @brief Base class for tasks.
     * To start a task, call kbf::task::start<MyTask>() where MyTask is a subclass of Task.
     */
    class Task {
        template<class T>
        friend std::shared_ptr<Task>
        start(const std::string &name, void *arg,  // NOLINT(readability-redundant-declaration
              uint32_t stackSize, uint32_t priority);

    public:
        /** @brief Name of the RTOS task. */
        const std::string name;

        /** @brief Task stack size. Default is 2048. */
        [[maybe_unused]] const uint32_t stackSize;

        /** @brief Task priority. Default is 5. */
        [[maybe_unused]] const uint32_t priority;

        /** @brief Tag used for logging. */
        static constexpr const char *const TAG = "kbf::task::Task";

        Task() = delete;

        /**
         * @brief Returns whether the task is running or not.
         * @return true if running, false otherwise
         */
        [[nodiscard]] bool running() const { return _running; }

        /**
         * @brief Stops the task.
         */
        void stop();

    protected:
        /**
         * @brief Constructor.
         * @param name name of the RTOS task
         * @param arg arguments to be passed to the task
         * @param stackSize stack size
         * @param priority priority
         */
        explicit Task(std::string name, void *arg, uint32_t stackSize, uint32_t priority);

        /**
         * @brief Function to be called when the task is started.
         * @param arg arguments
         */
        virtual void run(void *arg) = 0;

        /**
         * @brief Destructor.
         */
        ~Task();

    private:
        const uint32_t id;

        [[maybe_unused]] static void runInternal(void *arg);

        void *arg;

        TaskHandle_t handle{};

        bool _running = false;

        class TaskList {
            static constexpr const char *const TAG = "kbf::task::TaskList";
            template<class T>
            friend std::shared_ptr<Task>
            start(const std::string &name, void *arg,  // NOLINT(readability-redundant-declaration
                  uint32_t stackSize, uint32_t priority);

            friend class Task;

            static std::vector<std::shared_ptr<Task>> tasks;

            static void add(const std::shared_ptr<Task> &task);

            static void remove(uint32_t id);
        };
    };

    /**
     * @brief Starts a task.
     * @tparam T must be subclass of kbf::task::Task
     * @param name RTOS task name
     * @param arg arguments passed to run()
     * @param stackSize stack size
     * @param priority priority
     * @return shared_ptr to the newly created task
     */
    template<class T>
    std::shared_ptr<Task>
    start(const std::string &name, void *arg, const uint32_t stackSize, const uint32_t priority) {
        static_assert(std::is_base_of<Task, T>::value, "kbf::task::run(): type must be a kbf::task::Task");
        auto task = std::make_shared<T>(name, arg, stackSize, priority);

        esp_err_t err = xTaskCreate(Task::runInternal, name.c_str(), stackSize, task.get(), priority, &task->handle);
        if (err == errCOULD_NOT_ALLOCATE_REQUIRED_MEMORY) {
            throw kbf::exception::OutOfMemoryError();
        } else if (err != pdPASS) {
            throw kbf::exception::RTOSError(err);
        }

        Task::TaskList::add(task);
        return task;
    }
}

#endif //KBF_TASK_H
