#include "kbf/wifi.h"

#include <esp_log.h>
#include <esp_wifi.h>

#define KBF_INTERNAL

#include "kbf/internal/wifi.h"
#include "kbf/macros.h"
#include "kbf/exception.h"
#include "kbf/rtos.h"

using namespace kbf;
using namespace kbf::wifi;
using std::string;
using std::vector;

constexpr const char *const TAG = "kbf::wifi::sta";

namespace {
    string ssid;
    string password;

    int retryNum                           = 0;
    int retryMax                           = 0;

    bool         connected;
    kbf::net::IP ip; // NOLINT(cert-err58-cpp)

    kbf::rtos::EventGroup event; // NOLINT(cert-err58-cpp)
    const int             CONNECT_FINISHED = 0;

    esp_netif_obj     *interface = nullptr;
    wifi_sta_config_t staConfig  = {};

    esp_event_handler_instance_t connectHandler    = {};
    esp_event_handler_instance_t ipHandler         = {};
    esp_event_handler_instance_t disconnectHandler = {};
    esp_event_handler_instance_t scanDoneHandler   = {};

    void handleConnect(void *, esp_event_base_t, int32_t, void *);

    void handleDisconnect(void *, esp_event_base_t, int32_t, void *);

    void handleIp(void *, esp_event_base_t, int32_t, void *);

    void handleScanDone(void *, esp_event_base_t, int32_t, void *);

    void registerEventHandlers();

    void unregisterEventHandlers();
}

bool sta::connected() { return ::connected; }

const string &sta::ssid() { return ::ssid; }

const string &sta::password() { return ::password; }

const net::IP &sta::ip() { return ::ip; }

void (*sta::onConnect)() = nullptr;

void (*sta::onIp)() = nullptr;

void (*sta::onDisconnect)() = nullptr;

void (*sta::onScanDone)(vector<APInfo> &apList, void *data) = nullptr;

void sta::start() {
    ESP_LOGI(TAG, "%s()", __func__);

    if (internal::mode == Mode::STA || internal::mode == Mode::DUAL) throw exception::InvalidMode();

    internal::init();
    interface = esp_netif_create_default_wifi_sta();
    registerEventHandlers();

    staConfig.pmf_cfg = {true, false};
    wifi_config_t config = {.sta = staConfig};

    if (internal::mode == Mode::OFF) {
        CHECK(esp_wifi_set_mode(WIFI_MODE_STA));
        CHECK(esp_wifi_set_config(WIFI_IF_STA, &config));
        CHECK(esp_wifi_start());
        internal::setMode(Mode::STA);
    } else {
        CHECK(esp_wifi_set_mode(WIFI_MODE_APSTA));
        CHECK(esp_wifi_set_config(WIFI_IF_STA, &config));
        internal::setMode(Mode::DUAL);
    }
}

void sta::stop() {
    ESP_LOGI(TAG, "%s()", __func__);

    if (internal::mode != Mode::STA && internal::mode != Mode::DUAL) throw wifi::exception::InvalidMode();

    unregisterEventHandlers();

    if (internal::mode == Mode::DUAL) {
        CHECK(esp_wifi_set_mode(WIFI_MODE_AP));
        internal::setMode(Mode::AP);
    } else {
        CHECK(esp_wifi_stop());
        internal::setMode(Mode::OFF);
    }

    ::connected = false;
    esp_netif_destroy(interface);
}

bool sta::connect(string ssid, string password, bool async, int maxAttempts) {
    ESP_LOGI(TAG, "%s(%s, <omitted>, %d, %d)", __func__, ssid.c_str(), async, maxAttempts);

    if (internal::mode != Mode::STA && internal::mode != Mode::DUAL) throw wifi::exception::InvalidMode();

    retryNum = 0;
    retryMax = maxAttempts;

    event.clear();

    ::ssid     = std::move(ssid);
    ::password = std::move(password);

    std::copy(::ssid.begin(), ::ssid.end() + 1, std::begin(staConfig.ssid));
    std::copy(::password.begin(), ::password.end() + 1, std::begin(staConfig.password));

    wifi_config_t config = {.sta = staConfig};
    CHECK(esp_wifi_set_config(WIFI_IF_STA, &config));
    CHECK(esp_wifi_connect());

    if (!async) {
        // TODO KBF-35: handle the lack of a DHCP server on the AP
        event.waitForBit(CONNECT_FINISHED);
        return ::connected;
    }

    return true;
}

void sta::disconnect() {
    ESP_LOGI(TAG, "%s()", __func__);

    retryMax    = 0;
    ::connected = false;
    CHECK(esp_wifi_disconnect());
}

void sta::startScan(void *data) {
    ESP_LOGI(TAG, "%s()", __func__);

    if (!onScanDone) { ESP_LOGW(TAG, "%s: onScanDone not set", __func__); }

    CHECK(esp_event_handler_instance_register(WIFI_EVENT, WIFI_EVENT_SCAN_DONE, &handleScanDone, data,
                                              &scanDoneHandler));

    CHECK(esp_wifi_scan_start(nullptr, false));
}

namespace {
    void registerEventHandlers() {
        ESP_LOGI(TAG, "%s()", __func__);

        CHECK(esp_event_handler_instance_register(
                WIFI_EVENT, WIFI_EVENT_STA_CONNECTED, &handleConnect, nullptr, &connectHandler));
        CHECK(esp_event_handler_instance_register(
                WIFI_EVENT, WIFI_EVENT_STA_DISCONNECTED, &handleDisconnect, nullptr, &disconnectHandler));
        CHECK(esp_event_handler_instance_register(
                IP_EVENT, IP_EVENT_STA_GOT_IP, &handleIp, nullptr, &ipHandler));
    }

    void unregisterEventHandlers() {
        ESP_LOGI(TAG, "%s()", __func__);

        CHECK(esp_event_handler_instance_unregister(WIFI_EVENT, WIFI_EVENT_STA_CONNECTED, &connectHandler));
        CHECK(esp_event_handler_instance_unregister(WIFI_EVENT, WIFI_EVENT_STA_DISCONNECTED, &disconnectHandler));
        CHECK(esp_event_handler_instance_unregister(IP_EVENT, IP_EVENT_STA_GOT_IP, &ipHandler));

        // TODO these are supposed to be deprecated and should not be necessary
        // however, handleDisconnect will be called on wifi stop unless unregistered; IDF bug?
        CHECK(esp_event_handler_unregister(WIFI_EVENT, WIFI_EVENT_STA_CONNECTED, &handleConnect));
        CHECK(esp_event_handler_unregister(WIFI_EVENT, WIFI_EVENT_STA_DISCONNECTED, &handleDisconnect));
        CHECK(esp_event_handler_unregister(IP_EVENT, IP_EVENT_STA_GOT_IP, &handleIp));
    }

    void handleConnect(void *, esp_event_base_t, int32_t, void *) {
        ESP_LOGI(TAG, "%s()", __func__);

        ::connected = true;
        if (sta::onConnect) sta::onConnect();
    }

    void handleDisconnect(void *, esp_event_base_t, int32_t, void *) {
        ESP_LOGI(TAG, "%s()", __func__);

        if (retryNum++ < retryMax) {
            ESP_LOGI(TAG, "reconnecting, attempt %d / %d", retryNum, retryMax);
            CHECK(esp_wifi_connect());
        } else {
            ESP_LOGI(TAG, "disconnected");
            ::connected = false;
            if (sta::onDisconnect) sta::onDisconnect();
            event.setBit(CONNECT_FINISHED);
        }
    }

    void handleIp(void *, esp_event_base_t, int32_t, void *data) {
        ESP_LOGI(TAG, "%s()", __func__);

        auto eventData = static_cast<ip_event_got_ip_t *>(data);
        ::ip = net::IP(eventData->ip_info.ip);

        if (sta::onIp) sta::onIp();
        retryNum = 0;

        event.setBit(CONNECT_FINISHED);
    }

    void handleScanDone(void *data, esp_event_base_t, int32_t, void *) {
        ESP_LOGI(TAG, "%s()", __func__);

        CHECK(esp_event_handler_instance_unregister(WIFI_EVENT, WIFI_EVENT_SCAN_DONE, &scanDoneHandler));
        CHECK(esp_event_handler_unregister(WIFI_EVENT, WIFI_EVENT_SCAN_DONE, &::handleScanDone));

        if (!sta::onScanDone) {
            ESP_LOGW(TAG, "%s: onScanDone not set", __func__);
            return;
        }

        uint16_t size;
        esp_wifi_scan_get_ap_num(&size);
        wifi_ap_record_t apRecords[size];
        esp_wifi_scan_get_ap_records(&size, apRecords);

        vector<APInfo> result;

        for (int i = 0; i < size; i++) {
            APInfo apInfo = {reinterpret_cast<char *>(apRecords[i].ssid), apRecords[i].rssi};
            result.push_back(apInfo);
        }

        sta::onScanDone(result, data);
    }
}
